package speiger.src.coreengine.rendering.gui.font.providers;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import org.lwjgl.stb.STBTTFontinfo;
import org.lwjgl.stb.STBTruetype;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.system.MemoryUtil;

import speiger.src.collections.ints.sets.IntOpenHashSet;
import speiger.src.collections.ints.sets.IntSet;
import speiger.src.coreengine.assets.AssetLocation;
import speiger.src.coreengine.assets.base.IAsset;
import speiger.src.coreengine.assets.base.IAssetProvider;
import speiger.src.coreengine.assets.parsers.NativeMemoryParser;
import speiger.src.coreengine.rendering.gui.font.FontTexture;
import speiger.src.coreengine.rendering.gui.font.glyth.Glyth;
import speiger.src.coreengine.rendering.gui.font.glyth.GlythData;
import speiger.src.coreengine.rendering.gui.font.glyth.IGlythSheetInfo;
import speiger.src.coreengine.rendering.textures.custom.Drawable;

public class STBTrueTypeProvider implements IFontProvider {
	long data;
	STBTTFontinfo info;
	IntSet skip = new IntOpenHashSet();
	float oversample;
	float xOff;
	float yOff;
	float pointScale;
	float ascent;
	
	public STBTrueTypeProvider(long data, STBTTFontinfo info, float size, float oversample, float xOff, float yOff, String skip) {
		this.data = data;
		this.info = info;
		skip.codePoints().forEach(this.skip::add);
		this.oversample = oversample;
		this.xOff = xOff;
		this.yOff = yOff;
		pointScale = STBTruetype.stbtt_ScaleForPixelHeight(info, size * oversample);
		int[] ascent = new int[1];
		STBTruetype.stbtt_GetFontVMetrics(info, ascent, new int[1], new int[1]);
		this.ascent = pointScale * ascent[0];
	}
	
	public static IFontProvider create(AssetLocation location, IAssetProvider provider) {
		try(IAsset asset = provider.getAsset(location)) {
			ByteBuffer buffer = asset.custom(NativeMemoryParser.INSTANCE);
			STBTTFontinfo info = STBTTFontinfo.create();
			if(!STBTruetype.stbtt_InitFont(info, buffer)) {
				System.out.println("Couldn't load font");
				MemoryUtil.memFree(buffer);
				info.free();
				return null;
			}
			return new STBTrueTypeProvider(MemoryUtil.memAddress(buffer), info, 10.2F, 24F, 0, 0, "");
		}
		catch(Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	@Override
	public GlythData glythData(int codepoint, boolean bold) {
		if(skip.contains(codepoint)) return null;
		int glyth = STBTruetype.nstbtt_FindGlyphIndex(info.address(), codepoint);
		if(glyth == 0) return null;
		try(MemoryStack stack = MemoryStack.stackPush()) {
			IntBuffer left = stack.mallocInt(1);
			IntBuffer bottom = stack.mallocInt(1);
			IntBuffer right = stack.mallocInt(1);
			IntBuffer top = stack.mallocInt(1);
			IntBuffer advance = stack.mallocInt(1);
			IntBuffer leftSideBearing = stack.mallocInt(1);
			STBTruetype.stbtt_GetGlyphHMetrics(info, glyth, advance, leftSideBearing);
			STBTruetype.stbtt_GetGlyphBitmapBoxSubpixel(info, glyth, pointScale, pointScale, xOff, yOff, left, bottom, right, top);
			int minX = left.get(0);
			int minY = -top.get(0);
			int maxX = right.get(0);
			int maxY = -bottom.get(0);
			if(maxX - minX <= 0 || maxY - minY <= 0) return null;
			return new STBGlyth(this, minX, minY, maxX, maxY, advance.get(0) * pointScale, leftSideBearing.get(0) * pointScale, glyth);
		}
	}

	@Override
	public void close() {
		if(data == 0L) return;
		info.free();
		MemoryUtil.nmemFree(data);
		data = 0L;
	}
	
	private static class STBGlyth implements GlythData {
		final STBTrueTypeProvider owner;
		final int width;
		final int height;
//		final float xOff;
//		final float yOff;
		final float advance;
		final int glyth;
		
		public STBGlyth(STBTrueTypeProvider owner, int minX, int minY, int maxX, int maxY, float advance, float leftPadding, int glyth) {
			this.owner = owner;
			this.width = maxX - minX;
			this.height = maxY - minY;
			this.advance = advance / owner.oversample;
//			this.xOff = (leftPadding + minX + owner.xOff) / owner.oversample;
//			this.yOff = (owner.ascent + maxY + owner.yOff) / owner.oversample;
			this.glyth = glyth;
		}

		@Override
		public float advance() {
			return advance;
		}

		@Override
		public Glyth bake(GlythBaker baker) {
			return baker.bake(new IGlythSheetInfo() {
				@Override
				public int width() { return width; }
				@Override
				public int height() { return height; }
				@Override
				public boolean isColored() { return false; }
				@Override
				public void upload(int x, int y) {
					Drawable drawable = new Drawable(FontTexture.formatByColor(false), width, height);
					drawable.drawFont(owner.info, glyth, owner.xOff, owner.yOff, 0, 0, width, height, owner.pointScale, owner.pointScale);
					drawable.upload(x, y, 0, 0, width, height);
					drawable.close();
				}
			});
		}
		
	}
}
